import json
import logging
import pathlib
import threading
from typing import Any

from flask import Flask, request

from wi1_bot import push
from wi1_bot.arr import Radarr, Sonarr, replace_remote_paths
from wi1_bot.config import config
from wi1_bot.transcoder.transcode_queue import queue

app = Flask(__name__)

logging.getLogger("werkzeug").disabled = True

logger = logging.getLogger(__name__)

radarr = Radarr(config["radarr"]["url"], config["radarr"]["api_key"])
sonarr = Sonarr(config["sonarr"]["url"], config["sonarr"]["api_key"])


def on_grab(req: dict[str, Any]) -> None:
    push.send(req["release"]["releaseTitle"], title=f"file grabbed ({req['downloadClient']})")


def on_download(req: dict[str, Any]) -> None:
    path: pathlib.Path

    if "movie" in req:
        movie_json = radarr._radarr.get_movie(req["movie"]["id"])
        assert isinstance(movie_json, dict)

        quality_profile = radarr.get_quality_profile_name(movie_json["qualityProfileId"])

        movie_folder = req["movie"]["folderPath"]
        relative_path = req["movieFile"]["relativePath"]

        path = pathlib.Path(movie_folder) / relative_path

        if not req["isUpgrade"]:
            push.send(path.name, title="new movie downloaded")
    elif "series" in req:
        series_json = sonarr._sonarr.get_series(req["series"]["id"])
        assert isinstance(series_json, dict)

        quality_profile = sonarr.get_quality_profile_name(series_json["qualityProfileId"])

        series_folder = req["series"]["path"]
        relative_path = req["episodeFile"]["relativePath"]

        path = pathlib.Path(series_folder) / relative_path

        if not req["isUpgrade"]:
            push.send(path.name, title="new episode downloaded")
    else:
        raise ValueError("unknown download request")

    if "transcoding" not in config:
        return

    path = replace_remote_paths(path)

    quality_options = config["transcoding"]["profiles"][quality_profile]

    languages = quality_options.get("languages", None)
    video_params = quality_options.get("video_params", None)
    audio_params = quality_options.get("audio_params", None)

    queue.add(
        path=str(path),
        languages=languages,
        video_params=video_params,
        audio_params=audio_params,
    )


@app.route("/", methods=["POST"])
def index() -> Any:
    try:
        if request.json is None or "eventType" not in request.json:
            return "", 400

        logger.debug(f"got request: {json.dumps(request.json)}")

        if request.json["eventType"] == "Download":
            on_download(request.json)
    except Exception:
        logger.warning(f"error handling request: {request.data.decode()}", exc_info=True)

    return "", 200


@app.route("/health", methods=["GET"])
def health() -> Any:
    return "OK", 200


def start() -> None:
    logger.info("starting webhook listener")

    t = threading.Thread(target=app.run, kwargs={"host": "0.0.0.0", "port": 9000})
    t.daemon = True
    t.start()


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=9000)



from typing import Any


class Episode:
    def __init__(
        self,
        ep_json: dict[str, Any],
        *,
        series_title: str,
        series_tvdb_id: int,
        series_imdb_id: str,
    ) -> None:
        self._json = ep_json

        self.series_title: str = series_title
        self.season_num: int = ep_json["seasonNumber"]
        self.ep_num: int = ep_json["episodeNumber"]
        self.ep_title: str = ep_json["title"]
        self.air_date: str = ep_json["airDate"]
        self.tvdb_id: int = series_tvdb_id
        self.imdb_id: str = series_imdb_id

        self.full_title: str = (
            f"{self.series_title} S{self.season_num:02d}E{self.ep_num:02d} - {self.ep_title}"
        )

        if self.imdb_id:
            self.url = f"https://www.imdb.com/title/{self.imdb_id}"
        else:
            self.url = f"https://www.thetvdb.com/?id={self.tvdb_id}"

    def __str__(self) -> str:
        return f"[{self.full_title}]({self.url})"

    def __repr__(self) -> str:
        return str(self.__dict__)



from shutil import rmtree
from typing import Any

from pyarr import SonarrAPI

from .download import Download


class Series:
    def __init__(self, series_json: dict[str, Any]) -> None:
        self._json = series_json

        self.title: str = series_json["title"]
        self.year: int = series_json["year"]
        self.tvdb_id: int = series_json["tvdbId"]

        self.db_id: int | None = series_json["id"] if "id" in series_json else None

        self.full_title = f"{self.title} ({self.year})"

        self.url = f"https://thetvdb.com/dereferrer/series/{self.tvdb_id}"

        self.imdb_id = ""

        try:
            self.imdb_id = series_json["imdbId"]
            self.url = f"https://imdb.com/title/{self.imdb_id}"
        except KeyError:
            pass

    def __str__(self) -> str:
        return f"[{self.full_title}]({self.url})"

    def __repr__(self) -> str:
        return str(self.__dict__)


class SonarrError(Exception):
    pass


class Sonarr:
    def __init__(self, url: str, api_key: str) -> None:
        self._sonarr = SonarrAPI(url, api_key)

    def lookup_series(self, query: str) -> list[Series]:
        possible_series = self._sonarr.lookup_series(query)

        if isinstance(possible_series, dict):
            raise SonarrError(possible_series["message"])

        return [Series(s) for s in possible_series]

    def lookup_library(self, query: str) -> list[Series]:
        possible_series = self._sonarr.lookup_series(query)

        return [Series(s) for s in possible_series if "id" in s]

    def lookup_user_library(self, query: str, user_id: int) -> list[Series]:
        try:
            tag_id = self._get_tag_for_user_id(user_id)
        except ValueError:
            return []

        possible_series = self._sonarr.lookup_series(query)

        # self._sonarr.get_tag_detail is broken/not supported in v1 sonarr API so have
        # to filter the tmdb lookup (probably slower but saves an API call)
        user_series = [s for s in possible_series if tag_id in s["tags"]]

        return [Series(s) for s in user_series]

    def add_series(self, series: Series, profile: str = "good") -> bool:
        if series.db_id is not None:
            return False

        quality_profile_id = self._get_quality_profile_id(profile)
        language_profile_id = self._sonarr.get_language_profile()[0]["id"]
        root_folder = self._sonarr.get_root_folder()
        assert isinstance(root_folder, list)
        root_folder_path = root_folder[0]["path"]

        series_json = self._sonarr.add_series(
            series._json,
            quality_profile_id=quality_profile_id,
            language_profile_id=language_profile_id,
            root_dir=root_folder_path,
            search_for_missing_episodes=True,
        )

        series.db_id = series_json["id"]

        return True

    def del_series(self, series: Series) -> None:
        if series.db_id is None:
            raise ValueError(f"{series} is not in the library")

        series_json = self._sonarr.get_series(series.db_id)
        assert isinstance(series_json, dict)

        self._sonarr.del_series(series.db_id, delete_files=True)

        try:
            rmtree(series_json["path"])
        except FileNotFoundError:
            pass

    def series_downloaded(self, series: Series) -> bool:
        if series.db_id is None:
            return False

        files = self._sonarr.get_episode_files_by_series_id(series.db_id)

        if files:
            return True

        return False

    def create_tag(self, tag: str) -> None:
        self._sonarr.create_tag(tag)

    def add_tag(self, series: Series, user_id: int) -> bool:
        try:
            tag_id = self._get_tag_for_user_id(user_id)
        except ValueError:
            # tag_id = self._radarr.create_tag(str(user_id))['id']

            return False

        series_json = self._sonarr.get_series(series.db_id)
        assert isinstance(series_json, dict)

        series_json["tags"].append(tag_id)

        self._sonarr.upd_series(series_json)

        return True

    def get_downloads(self) -> list[Download]:
        queue = self._sonarr.get_queue()["records"]

        downloads = [Download(d) for d in queue]

        return downloads

    def get_quota_amount(self, user_id: int) -> int:
        try:
            tag_id = self._get_tag_for_user_id(user_id)
        except ValueError:
            return 0

        total = 0

        for series in self._sonarr.get_series():
            assert isinstance(series, dict)
            if tag_id in series["tags"]:
                total += series["statistics"]["sizeOnDisk"]

        return total

    def get_quality_profile_name(self, profile_id: int) -> str:
        profiles = self._sonarr.get_quality_profile()

        for profile in profiles:
            if profile["id"] == profile_id:
                name = profile["name"]
                assert isinstance(name, str)
                return name

        raise ValueError(f"no quality profile with the id {profile_id}")

    def get_series(self) -> list[dict[str, Any]]:
        series = self._sonarr.get_series()
        assert isinstance(series, list)
        return series

    def rescan_series(self, series_id: int) -> None:
        self._sonarr.post_command("RescanSeries", seriesId=series_id)  # type: ignore

    def _get_quality_profile_id(self, name: str) -> int:
        profiles = self._sonarr.get_quality_profile()

        for profile in profiles:
            if profile["name"].lower() == name.lower():
                profile_id = profile["id"]
                assert isinstance(profile_id, int)
                return profile_id

        raise ValueError(f"no quality profile with the name {name}")

    def _get_tag_for_user_id(self, user_id: int) -> int:
        tags = self._sonarr.get_tag()
        assert isinstance(tags, list)

        for tag in tags:
            if str(user_id) in tag["label"]:
                tag_id = tag["id"]
                assert isinstance(tag_id, int)
                return tag_id

        raise ValueError(f"no tag with the user id {user_id}")



from shutil import rmtree
from typing import Any

from pyarr import RadarrAPI

from .download import Download
from .movie import Movie


class Radarr:
    def __init__(self, url: str, api_key: str) -> None:
        self._radarr = RadarrAPI(url, api_key)

    def lookup_movie(self, query: str) -> list[Movie]:
        possible_movies = self._radarr.lookup_movie(query)

        return [Movie(m) for m in possible_movies]

    def lookup_library(self, query: str) -> list[Movie]:
        possible_movies = self._radarr.lookup_movie(query)

        return [Movie(m) for m in possible_movies if "id" in m]

    def lookup_user_library(self, query: str, user_id: int) -> list[Movie]:
        try:
            tag_id = self._get_tag_for_user_id(user_id)
        except ValueError:
            return []

        tag_detail = self._radarr.get_tag_detail(tag_id)
        assert isinstance(tag_detail, dict)

        possible_movies = self._radarr.lookup_movie(query)

        user_movie_ids = tag_detail["movieIds"]

        return [Movie(m) for m in possible_movies if "id" in m and m["id"] in user_movie_ids]

    def add_movie(self, movie: Movie, profile: str = "good") -> bool:
        if self._radarr.get_movie(movie.tmdb_id, tmdb=True):
            return False

        quality_profile_id = self._get_quality_profile_id(profile)

        root_folder = self._radarr.get_root_folder()
        assert isinstance(root_folder, list)
        root_folder_path = root_folder[0]["path"]

        self._radarr.add_movie(
            movie._json,
            root_dir=root_folder_path,
            quality_profile_id=quality_profile_id,
        )

        return True

    def del_movie(self, movie: Movie) -> None:
        potential = self._radarr.get_movie(movie.tmdb_id, tmdb=True)
        assert isinstance(potential, list)

        if not potential:
            raise ValueError(f"{movie} is not in the library")

        movie_json = potential[0]

        db_id = movie_json["id"]
        path = movie_json["folderName"]

        self._radarr.del_movie(db_id, delete_files=True, add_exclusion=False)

        try:
            rmtree(path)
        except FileNotFoundError:
            pass

    def movie_downloaded(self, movie: Movie) -> bool:
        potential = self._radarr.get_movie(movie.tmdb_id, tmdb=True)
        assert isinstance(potential, list)

        if not potential:
            return False

        files = self._radarr.get_movie_files_by_movie_id(potential[0]["id"])

        return len(files) > 0

    def create_tag(self, tag: str) -> None:
        self._radarr.create_tag(tag)

    def add_tag(self, movie: Movie | list[Movie], user_id: int) -> bool:
        if isinstance(movie, Movie):
            json = self._radarr.get_movie(movie.tmdb_id, tmdb=True)
            assert isinstance(json, list)
            ids = [json[0]["id"]]
        else:
            ids = []

            for m in movie:
                json = self._radarr.get_movie(m.tmdb_id, tmdb=True)
                assert isinstance(json, list)
                ids.append(json[0]["id"])

        try:
            tag_id = self._get_tag_for_user_id(user_id)
        except ValueError:
            # tag_id = self._radarr.create_tag(str(user_id))['id']

            return False

        edit_json = {"movieIds": ids, "tags": [tag_id], "applyTags": "add"}

        self._radarr.upd_movies(edit_json)

        return True

    def get_downloads(self) -> list[Download]:
        queue = self._radarr.get_queue_details()

        downloads = [Download(d) for d in queue]

        return sorted(downloads, key=lambda d: (d.timeleft, -d.pct_done))

    def get_quota_amount(self, user_id: int) -> int:
        try:
            tag_id = self._get_tag_for_user_id(user_id)
        except ValueError:
            return 0

        total = 0

        for movie in self._radarr.get_movie():
            assert isinstance(movie, dict)

            if tag_id in movie["tags"]:
                total += movie["sizeOnDisk"]

        return total

    def get_quality_profile_name(self, profile_id: int) -> str:
        profiles = self._radarr.get_quality_profile()

        for profile in profiles:
            if profile["id"] == profile_id:
                name = profile["name"]
                assert isinstance(name, str)
                return name

        raise ValueError(f"no quality profile with the id {profile_id}")

    def get_movies(self) -> list[dict[str, Any]]:
        movies = self._radarr.get_movie()
        assert isinstance(movies, list)
        return movies

    def rescan_movie(self, movie_id: int) -> None:
        self._radarr.post_command("RescanMovie", movieId=movie_id)  # type: ignore

    def refresh_movie(self, movie_id: int) -> None:
        self._radarr.post_command("RefreshMovie", movieIds=[movie_id])  # type: ignore

    def search_missing(self) -> None:
        self._radarr.post_command(name="MissingMoviesSearch")

    def _get_quality_profile_id(self, name: str) -> int:
        profiles = self._radarr.get_quality_profile()

        for profile in profiles:
            if profile["name"].lower() == name.lower():
                profile_id = profile["id"]
                assert isinstance(profile_id, int)
                return profile_id

        raise ValueError(f"no quality profile with the name {name}")

    def _get_tag_for_user_id(self, user_id: int) -> int:
        tags = self._radarr.get_tag()
        assert isinstance(tags, list)

        for tag in tags:
            if str(user_id) in tag["label"]:
                tag_id = tag["id"]
                assert isinstance(tag_id, int)
                return tag_id

        raise ValueError(f"no tag with the user id {user_id}")


__all__ = ["Movie"]


if __name__ == "__main__":
    from wi1_bot.config import config

    radarr = Radarr(config["radarr"]["url"], config["radarr"]["api_key"])
    pyarr = radarr._radarr
